//******************************************************************************
// Designer     : Akashi Satoh
// Emp No.      :
// E-Mail       :
// Dept         :
//
// Version      : 0.1
// Created      : Sep/8/2009
// Last Updated : Sep/8/2009
//
// File Name    : aes_multi_core.v
//******************************************************************************
`timescale 1ns / 1ps

// =====================================================================================================
//      GF(2^2^2^2) inverter
// =====================================================================================================
module GFinv(x, y);
  input  [7:0] x;
  output [7:0] y;

  wire [8:0] da, db, dx, dy, va, tp, tn;
  wire [3:0] u, v;
  wire [4:0] mx;
  wire [5:0] my;

// Equetions
  assign da ={x[3], x[2]^x[3], x[2], x[1]^x[3], x[0]^x[1]^x[2]^x[3], x[0]^x[2], x[1], x[0]^x[1], x[0]};
  assign db ={x[7], x[6]^x[7], x[6], x[5]^x[7], x[4]^x[5]^x[6]^x[7], x[4]^x[6], x[5], x[4]^x[5], x[4]};
  assign va ={v[3], v[2]^v[3], v[2], v[1]^v[3], v[0]^v[1]^v[2]^v[3], v[0]^v[2], v[1], v[0]^v[1], v[0]};
  assign dx = da ^ db;
  assign dy = da & dx;
  assign tp = va & dx;
  assign tn = va & db;

  assign u = {dy[0] ^ dy[1] ^ dy[3] ^ dy[4] ^ x[4] ^ x[5] ^ x[6],
              dy[0] ^ dy[2] ^ dy[3] ^ dy[5] ^ x[4] ^ x[7],
              dy[0] ^ dy[1] ^ dy[7] ^ dy[8] ^ x[7],
              dy[0] ^ dy[2] ^ dy[6] ^ dy[7] ^ x[6] ^ x[7]};

  assign y = {tn[0] ^ tn[1] ^ tn[3] ^ tn[4], tn[0] ^ tn[2] ^ tn[3] ^ tn[5],
              tn[0] ^ tn[1] ^ tn[7] ^ tn[8], tn[0] ^ tn[2] ^ tn[6] ^ tn[7],
              tp[0] ^ tp[1] ^ tp[3] ^ tp[4], tp[0] ^ tp[2] ^ tp[3] ^ tp[5],
              tp[0] ^ tp[1] ^ tp[7] ^ tp[8], tp[0] ^ tp[2] ^ tp[6] ^ tp[7]};

  // GF(2^2^2) Inverter 
  assign mx = { mx[0] ^ mx[1] ^ u[2],
                mx[0] ^ mx[2] ^ u[3],
                u[1] & (u[1] ^ u[3]),
                ( u[0] ^ u[1]) & (u[0] ^ u[1]  ^ u[2] ^ u[3] ),
                u[0] & (u[0] ^ u[2]) };

  assign my = { ~( mx[4] & u[3] ),
                ~( mx[3] & (u[2] ^ u[3]) ),
                ~( (mx[3] ^ mx[4]) & u[2] ),
                ~( mx[4] & (u[1] ^ u[3]) ),
                ~( mx[3] & (u[0] ^ u[1]  ^ u[2] ^ u[3]) ),
                ~( (mx[3] ^ mx[4]) & (u[0] ^ u[2])) };

  assign v = { my[3] ^ my[4], my[3] ^ my[5], my[0] ^ my[1], my[0] ^ my[2] };
endmodule

// ===============================================================================
// Sbox GF(2^2^2^2)
// ===============================================================================
module SboxComp(x, y);
  input  [7:0] x;
  output [7:0] y;

  wire [7:0] a, b;

  assign a = {x[5] ^ x[7],
              x[1] ^ x[2] ^ x[3] ^ x[4] ^ x[6] ^ x[7],
              x[2] ^ x[3] ^ x[5] ^ x[7],
              x[1] ^ x[2] ^ x[3] ^ x[5] ^ x[7],
              x[1] ^ x[2] ^ x[6] ^ x[7],
              x[1] ^ x[2] ^ x[3] ^ x[4] ^ x[7],
              x[1] ^ x[4] ^ x[6],
              x[0] ^ x[1] ^ x[6]};

  GFinv GFinv(a, b);

  assign y = { b[2] ^ b[3] ^ b[7],
              ~b[4] ^ b[5] ^ b[6] ^ b[7],
              ~b[2] ^ b[7],
               b[0] ^ b[1] ^ b[4] ^ b[7],
               b[0] ^ b[1] ^ b[2],
               b[0] ^ b[2] ^ b[3] ^ b[4] ^ b[5] ^ b[6],
              ~b[0] ^ b[7],
              ~b[0] ^ b[1] ^ b[2] ^ b[6] ^ b[7]};
endmodule

// ===============================================================================
// InvSbox GF(2^2^2^2)
// ===============================================================================
module InvSboxComp(x, y);
  input  [7:0] x;
  output [7:0] y;

  wire [7:0] a, b;

  assign a = { x[1] ^ x[2] ^ x[6] ^ x[7],
              ~x[0] ^ x[1] ^ x[2] ^ x[3] ^ x[6] ^ x[7],
              ~x[0] ^ x[4] ^ x[5] ^ x[6],
              ~x[3] ^ x[4] ^ x[5],
              ~x[5] ^ x[7],
              ~x[1] ^ x[2] ^ x[5] ^ x[6] ^ x[7],
               x[1] ^ x[3] ^ x[5],
              ~x[2] ^ x[6] ^ x[7]};

  GFinv GFinv(a, b);

  assign y = {b[1] ^ b[5] ^ b[6] ^ b[7],
              b[2] ^ b[6],
              b[1] ^ b[5] ^ b[6],
              b[1] ^ b[2] ^ b[4] ^ b[5] ^ b[6],
              b[1] ^ b[2] ^ b[3] ^ b[4] ^ b[5],
              b[1] ^ b[2] ^ b[3] ^ b[4] ^ b[7],
              b[4] ^ b[5],
              b[0] ^ b[2] ^ b[4] ^ b[5] ^ b[6]};
endmodule

// ===============================================================================
// SubBytes GF(2^2^2^2)
// ===============================================================================
module SubBytes (x, y);
  input  [31:0] x;
  output [31:0] y;

  SboxComp Sbox3(x[31:24], y[31:24]);
  SboxComp Sbox2(x[23:16], y[23:16]);
  SboxComp Sbox1(x[15: 8], y[15: 8]);
  SboxComp Sbox0(x[ 7: 0], y[ 7: 0]);
endmodule

// ===============================================================================
// InvSubBytes GF(2^2^2^2)
// ===============================================================================
module InvSubBytes (x, y);
  input  [31:0] x;
  output [31:0] y;

  InvSboxComp Sbox3(x[31:24], y[31:24]);
  InvSboxComp Sbox2(x[23:16], y[23:16]);
  InvSboxComp Sbox1(x[15: 8], y[15: 8]);
  InvSboxComp Sbox0(x[ 7: 0], y[ 7: 0]);
endmodule

// ===============================================================================
// MixColumns
// ===============================================================================
module MixColumns(x, y);
  input  [31:0]  x;
  output [31:0]  y;

  wire [7:0] a3, a2, a1, a0, b3, b2, b1, b0;

  assign a3 = x[31:24]; assign a2 = x[23:16];
  assign a1 = x[15: 8]; assign a0 = x[ 7: 0];

  assign b3 = a3 ^ a2; assign b2 = a2 ^ a1;
  assign b1 = a1 ^ a0; assign b0 = a0 ^ a3;

  assign y = {a2[7] ^ b1[7] ^ b3[6],         a2[6] ^ b1[6] ^ b3[5],
              a2[5] ^ b1[5] ^ b3[4],         a2[4] ^ b1[4] ^ b3[3] ^ b3[7],
              a2[3] ^ b1[3] ^ b3[2] ^ b3[7], a2[2] ^ b1[2] ^ b3[1],
              a2[1] ^ b1[1] ^ b3[0] ^ b3[7], a2[0] ^ b1[0] ^ b3[7],
              a3[7] ^ b1[7] ^ b2[6],         a3[6] ^ b1[6] ^ b2[5],
              a3[5] ^ b1[5] ^ b2[4],         a3[4] ^ b1[4] ^ b2[3] ^ b2[7],
              a3[3] ^ b1[3] ^ b2[2] ^ b2[7], a3[2] ^ b1[2] ^ b2[1],
              a3[1] ^ b1[1] ^ b2[0] ^ b2[7], a3[0] ^ b1[0] ^ b2[7],
              a0[7] ^ b3[7] ^ b1[6],         a0[6] ^ b3[6] ^ b1[5],
              a0[5] ^ b3[5] ^ b1[4],         a0[4] ^ b3[4] ^ b1[3] ^ b1[7],
              a0[3] ^ b3[3] ^ b1[2] ^ b1[7], a0[2] ^ b3[2] ^ b1[1],
              a0[1] ^ b3[1] ^ b1[0] ^ b1[7], a0[0] ^ b3[0] ^ b1[7],
              a1[7] ^ b3[7] ^ b0[6],         a1[6] ^ b3[6] ^ b0[5],
              a1[5] ^ b3[5] ^ b0[4],         a1[4] ^ b3[4] ^ b0[3] ^ b0[7],
              a1[3] ^ b3[3] ^ b0[2] ^ b0[7], a1[2] ^ b3[2] ^ b0[1],
              a1[1] ^ b3[1] ^ b0[0] ^ b0[7], a1[0] ^ b3[0] ^ b0[7]};
endmodule

// ===============================================================================
//  InvMixColumns
//  MixCol * {0,4,5}matrix version
// ===============================================================================
module InvMixColumns(x, y);
  input  [31:0]  x;
  output [31:0]  y;

  wire [7:0] a3, a2, a1, a0, b3, b2, b1, b0;
  wire [7:0] c3, c2, c1, c0, d3, d2, d1, d0;

  assign { a3, a2, a1, a0 } =  x;
  assign { b3, b2, b1,b0 } = { a3 ^ a2, a2 ^ a1, a1 ^ a0, a0 ^ a3 };

  assign c3 = {a2[7] ^ b1[7] ^ b3[6],         a2[6] ^ b1[6] ^ b3[5],
               a2[5] ^ b1[5] ^ b3[4],         a2[4] ^ b1[4] ^ b3[3] ^ b3[7],
               a2[3] ^ b1[3] ^ b3[2] ^ b3[7], a2[2] ^ b1[2] ^ b3[1],
               a2[1] ^ b1[1] ^ b3[0] ^ b3[7], a2[0] ^ b1[0] ^ b3[7]};
  assign c2 = {a3[7] ^ b1[7] ^ b2[6],         a3[6] ^ b1[6] ^ b2[5],
               a3[5] ^ b1[5] ^ b2[4],         a3[4] ^ b1[4] ^ b2[3] ^ b2[7],
               a3[3] ^ b1[3] ^ b2[2] ^ b2[7], a3[2] ^ b1[2] ^ b2[1],
               a3[1] ^ b1[1] ^ b2[0] ^ b2[7], a3[0] ^ b1[0] ^ b2[7]};
  assign c1 = {a0[7] ^ b3[7] ^ b1[6],         a0[6] ^ b3[6] ^ b1[5],
               a0[5] ^ b3[5] ^ b1[4],         a0[4] ^ b3[4] ^ b1[3] ^ b1[7],
               a0[3] ^ b3[3] ^ b1[2] ^ b1[7], a0[2] ^ b3[2] ^ b1[1],
               a0[1] ^ b3[1] ^ b1[0] ^ b1[7], a0[0] ^ b3[0] ^ b1[7]};
  assign c0 = {a1[7] ^ b3[7] ^ b0[6],         a1[6] ^ b3[6] ^ b0[5],
               a1[5] ^ b3[5] ^ b0[4],         a1[4] ^ b3[4] ^ b0[3] ^ b0[7],
               a1[3] ^ b3[3] ^ b0[2] ^ b0[7], a1[2] ^ b3[2] ^ b0[1],
               a1[1] ^ b3[1] ^ b0[0] ^ b0[7], a1[0] ^ b3[0] ^ b0[7]};
  assign d3 = {c3[5], c3[4], c3[3] ^ c3[7], c3[2] ^ c3[7] ^ c3[6],
               c3[1] ^ c3[6], c3[0] ^ c3[7], c3[7] ^ c3[6], c3[6]};
  assign d2 = {c2[5], c2[4], c2[3] ^ c2[7], c2[2] ^ c2[7] ^ c2[6],
               c2[1] ^ c2[6], c2[0] ^ c2[7], c2[7] ^ c2[6], c2[6]};
  assign d1 = {c1[5], c1[4], c1[3] ^ c1[7], c1[2] ^ c1[7] ^ c1[6],
               c1[1] ^ c1[6], c1[0] ^ c1[7], c1[7] ^ c1[6], c1[6]};
  assign d0 = {c0[5], c0[4], c0[3] ^ c0[7], c0[2] ^ c0[7] ^ c0[6],
               c0[1] ^ c0[6], c0[0] ^ c0[7], c0[7] ^ c0[6], c0[6]};
  assign y  = {d3 ^ d1 ^ c3, d2 ^ d0 ^ c2, d3 ^ d1 ^ c1, d2 ^ d0 ^ c0};
endmodule

// ===============================================================================
// ShiftRows
// ===============================================================================
module ShiftRows(x, y);
  input  [127:0] x;
  output [127:0] y;

  assign y = { x[127:120], x[ 87: 80], x[ 47: 40], x[  7:  0],
               x[ 95: 88], x[ 55: 48], x[ 15:  8], x[103: 96],
               x[ 63: 56], x[ 23: 16], x[111:104], x[ 71: 64],
               x[ 31: 24], x[119:112], x[ 79: 72], x[ 39: 32] };
endmodule

// ===============================================================================
// InvShiftRows
// ===============================================================================
module InvShiftRows(x, y);
  input  [127:0] x;
  output [127:0] y;

  assign y = { x[127:120], x[ 23: 16], x[ 47: 40], x[ 71: 64],
               x[ 95: 88], x[119:112], x[ 15:  8], x[ 39: 32],
               x[ 63: 56], x[ 87: 80], x[111:104], x[  7:  0],
               x[ 31: 24], x[ 55: 48], x[ 79: 72], x[103: 96]};
endmodule

// ===============================================================================
// Rcon
// ===============================================================================
module Rcon(x, inv, y);
  input  [7:0] x;
  input        inv;
  output [7:0] y;

  wire [7:0] t, ti;

  assign t =  { x[6], x[5], x[4], x[3] ^ x[7], x[2] ^ x[7], x[1], x[0] ^ x[7], x[7] };
  assign ti = { x[0], x[7], x[6], x[5], x[4] ^ x[0], x[3] ^ x[0], x[2], x[1] ^ x[0] };
  assign y = (inv == 0)? t: ti;
endmodule


// ===============================================================================
// Encryption Key Generator Unit for 128-/192-/256-bit keys Expansion
// ===============================================================================
module EncKeyGen( k128, k192, k256, key0, key1, key_init0, key_init1, mode, count, rc,
                  key_add, key_next0, key_next1, rc_next);

  input          k128, k192, k256;
  input  [127:0] key0, key1;                // Key 0-127bit, Key 128-256
  input  [127:0] key_init0, key_init1;      // Initial Key 0-127bit, Key 128-256
  input    [1:0] mode;                      // 00:128-bit 10:192-bit 1x:256bit
  input   [14:0] count;                     // Counter
  input    [7:0] rc;                        // rc/invrc
  output [127:0] key_add;                   // Current Round Key
  output [127:0] key_next0;                 // Next Key 0-127bit
  output [127:0] key_next1;                 // Next Key 128-256bit
  output   [7:0] rc_next;                   // Next rc

  wire    [14:0] c;                         // copy of count[14:0]
  wire           s192key0, s192key1;        // key generation for Low/High of 192bits
  wire           s256key0, s256key1;        // key generation for Low/High of 256bits    

  wire   [127:0] kin;
  wire   [127:0] ki;
  reg    [127:0] kp;
  reg     [31:0] ksX;
  wire    [31:0] rotate_key;
  wire    [31:0] sbox_in;
  wire    [31:0] sbox_out;
  wire     [7:0] rcx;
  wire     [7:0] rx;
  wire    [31:0] rcon_out;
  wire   [127:0] ko;
  wire   [127:0] key2, key3;
  wire           sel_round_key256;
  wire           sel_round_key192a;
  wire           sel_round_key192b;

// ---------------------------------
// Equetions
// ---------------------------------
  assign c = count;
  assign s192key0 = k192 & (c[0] | c[2] | c[3] | c[5] | c[6] | c[8] | c[9] | c[11]);
  assign s192key1 = k192 & (c[1] | c[2] | c[4] | c[5] | c[7] | c[8] | c[10]);
  assign s256key0 = k256 & (c[0] | c[2] | c[4] | c[6] | c[8] | c[10] | c[12] | c[14]);
  assign s256key1 = k256 & (c[1] | c[3] | c[5] | c[7] | c[9] | c[11] | c[13]);

  assign sel_nextkey0 = k128 | s192key0 | s256key0;
  assign sel_nextkey1 = c[0] | s192key1 | s256key1;
  Rcon Rcon(rc, 1'b0, rcx);
  assign rc_next = (sel_nextkey0 == 1)? rcx: rc;

  assign kin = (count[0] == 1)? key_init0 : key0;   // Round0 or other
  assign ki = (sel_nextkey0 == 1)? kin: key1;

  // Next key1
  always @( mode or key_init1 or key0 or key1 or c or ko ) begin
//    if ( count[0] == 1 ) kp = key_init1;
    if ( c[0] == 1 ) kp = key_init1;
    else begin
      if ( mode != 2'h1 ) kp[127:0] = ko[127:0];
      else begin
	     if (( c[2] | c[5] | c[8] ) == 1 ) begin
          kp[127:96] = ko[31:0] ^ key1[127:96];
          kp[ 95:64] = ko[31:0] ^ key1[127:96] ^ key1[95:64];
          kp[63:0] = ko[63:0];
        end
        else begin
          kp[127:96] = key0[31:0] ^ key1[127:96];
          kp[ 95:64] = key0[31:0] ^ key1[127:96] ^ key1[95:64];
          kp[63:0] = ko[63:0];
        end
      end
    end
  end

  always @( count or k128 or k256 or k192 or s256key1 or key_init1 or key0 or key1 or ki or ko ) begin
    if ( k128 == 1'b1 ) ksX = ki[31:0];
    else if ( k256 == 1'b1 )                           // 256-bit key
      if ( count[0] == 1'b1 ) ksX = key_init1[31:0];
      else if ( s256key1 == 1'b1 ) ksX = key0[31:0];
      else ksX = key1[31:0];
    else begin                                         // 192-bit key
      if ( count[0] == 1'b1 ) ksX = key_init1[95:64];
      else ksX = key1[95:64];
    end
  end

  assign rotate_key = { ksX[23:0], ksX[31:24] };    // Rotate key
  assign sbox_in = (sel_nextkey0 == 1)? rotate_key : ksX;

  SubBytes SBOX ( sbox_in, sbox_out );
  assign rx = (sel_nextkey0 == 1)? rc: 8'h00;
  assign rcon_out = { sbox_out[31:24] ^ ( rc & rx ), sbox_out[23:0] };

  assign ko[127:96] = ki[127:96] ^ rcon_out;
  assign ko[ 95: 0] = ki[ 95: 0] ^ ko[127:32];

  assign key_next0 = (sel_nextkey0 == 1)? ko: key0;
  assign key_next1 = (sel_nextkey1 == 1)? kp: key1;

  assign sel_round_key256  = k256 & (c[1] | c[3] | c[5] | c[7] | c[9] | c[11] | c[13]);  // 256-bit key only
  assign sel_round_key192a = k192 & (c[1] | c[4] | c[7] | c[10]);                        // key 4,5,0,1
  assign sel_round_key192b = k192 & (c[2] | c[5] | c[8] | c[11]);                        // key 2,3,4,5

  assign key2 = { key1[127:64], key0[127:64] };
  assign key3 = { key0[ 63: 0], key1[127:64] };

  assign key_add = (sel_round_key256  == 1)? key1:  // k3,k4,k5,k6
                   (sel_round_key192a == 1)? key2:  // k4,k5,k0,k1
                   (sel_round_key192b == 1)? key3:  // k2,k3,k4,k5
                                             key0;  // k0,k1,k2,k3

endmodule

// ===============================================================================
// Decryotion Key Generator Unit for 128-/192-/256-bit keys Expansion
// ===============================================================================
module DecKeyGen( k128, k192, k256, key0, key1, key_init0, key_init1, inv, mode, count, rc,
                  key_add, key_next0, key_next1, rc_next);

  input          k128, k192, k256;
  input  [127:0] key0, key1;                // Key 0-127bit, Key 128-256
  input  [127:0] key_init0, key_init1;      // Initial Key 0-127bit, Key 128-256
  input          inv;
  input    [1:0] mode;                      // 00:128-bit 10:192-bit 1x:256bit
  input   [14:0] count;                     // Counter
  input    [7:0] rc;                        // rc/invrc
  output [127:0] key_add;                   // Current Round Key
  output [127:0] key_next0;                 // Next Key 0-127bit
  output [127:0] key_next1;                 // Next Key 128-256bit
  output   [7:0] rc_next;                   // Next rc

  wire    [14:0] c;                         // copy of count[14:0]
  wire           s192key0, s192key1;        // key generation for Low/High of 192bits
  wire           s256key0, s256key1;        // key generation for Low/High of 256bits    

  wire   [127:0] kin;
  wire   [127:0] ki, kp, kpx;
  reg    [127:0] kp1;
  wire    [31:0] ksL, ksH, ksX, ksH0, ksH1;
  wire    [31:0] rotate_key;
  wire    [31:0] sbox_out;
  wire     [7:0] rcx;
  wire     [7:0] rx;
  wire    [31:0] rcon_out;
  wire   [127:0] ko;
  wire    [31:0] sbox_in;

  wire   [127:0] key2, key3;
  wire           sel_round_key256;
  wire           sel_round_key192a;
  wire           sel_round_key192b;
  wire   [127:0] key_addX;
  wire    [31:0] x4, x5, x6, x8;

// ---------------------------------
// Equetions
// ---------------------------------
  assign c = count;
  assign s192key0 = k192 & (c[0] | c[2] | c[3] | c[5] | c[6] | c[8] | c[9] | c[11]);
  assign s192key1 = k192 & (c[1] | c[2] | c[4] | c[5] | c[7] | c[8] | c[10]);
  assign s256key0 = k256 & (c[0] | c[2] | c[4] | c[6] | c[8] | c[10] | c[12] | c[14]);
  assign s256key1 = k256 & (c[1] | c[3] | c[5] | c[7] | c[9] | c[11] | c[13]);

  assign sel_nextkey0 = k128 | s192key0 | s256key0;
  assign sel_nextkey1 = s192key1 | s256key1 | c[0];
  Rcon Rcon(rc, 1'b1, rcx);
  assign rc_next = (sel_nextkey0 == 1)? rcx: rc;

  assign x4 = kin[ 31: 0] ^ kin[ 63:32];
  assign kin = (count[0] == 1)? key_init0 : key0;   // Round0 or other
  assign ki = (sel_nextkey0 == 1)? kin: key1;
//  assign kp = (count[0] == 1)? key_init1 : ko;   // Round0 or other

  assign kpx[127:96] = (mode != 2'h1)? ko[127:96]: x6;  // x6 when 192-bit key
  assign kpx[ 95:64] = (mode != 2'h1)? ko[ 95:64]: x8;  // x8 when 192-bit key
  assign kpx[ 63:32] = ko[ 63:32];
  assign kpx[ 31: 0] = ko[ 31: 0];
  assign kp          = (c[0] == 1)? key_init1: kpx;

  assign x5 = (c[2] | c[5] | c[8] == 1)? ko[ 31: 0]: key0[31:0];  // selection for 192-bit key
  assign x6 = x5 ^ key1[127:96];
  assign x8 = key1[127:96] ^ key1[ 95:64];  

  // Next key1
  always @( count or mode or key_init1 or key0 or key1 or c or ko ) begin
    if ( count[0] == 1 ) kp1 = key_init1;
    else begin
      if ( mode != 2'h1 ) kp1[127:64] = ko[127:64];
      else begin
	     if (( c[2] | c[5] | c[8] ) == 1 ) begin
          kp1[127:96] = ko[31:0] ^ key1[127:96];
          kp1[ 95:64] = ko[31:0] ^ key1[127:96] ^ key1[95:64];
        end
        else begin
          kp1[127:96] = key0[31:0] ^ key1[127:96];
          kp1[ 95:64] = key0[31:0] ^ key1[127:96] ^ key1[95:64];
        end
      end

      kp1[63:0] = ko[63:0];
    end
  end


  assign ksL  = (k256 == 1)? key0[ 31: 0] : x4;                    // 256-bit key or 128-bit key
  assign ksH0 = (k192 == 1)? key_init1[ 95:64]: key_init1[ 31: 0]; // 192bit?
  assign ksH1 = (k192 == 1)? key1[ 95:64]     : key1[ 31: 0];      // 192bit?
  assign ksH  = (c[0] == 1)? ksH0             : ksH1;              // round0?

//  assign ksH  = (c[0] == 1)? key_init1 : key1;              // round0? 128 or 256
  assign ksX  = (k128 == 1 || s256key1 == 1)? ksL: ksH;

  assign rotate_key = { ksX[23:0], ksX[31:24] };     // Rotate key
  assign sbox_in = (sel_nextkey0 == 1)? rotate_key : ksX;
  SubBytes SBOX ( sbox_in, sbox_out );

  assign rx = (sel_nextkey0 == 1)? rc: 8'h00;
  assign rcon_out = { sbox_out[31:24] ^ ( rc & rx ), sbox_out[23:0] };

  assign ko[127:96] = ki[127:96] ^ rcon_out;
  assign ko[ 95: 0] = ki[ 95: 0] ^ (( inv == 1'b1 )? ki[127:32] : ko[127:32] );

  assign key_next0 = (sel_nextkey0 == 1)? ko: key0;
  assign key_next1 = (sel_nextkey1 == 1)? kp: key1;

  assign sel_round_key256  = k256 & (c[1] | c[3] | c[5] | c[7] | c[9] | c[11] | c[13]);  // 256-bit key only
  assign sel_round_key192a = k192 & (c[2] | c[5] | c[8] | c[11]);
  assign sel_round_key192b = k192 & (c[1] | c[4] | c[7] | c[10]);

  assign key2 = { key1[127:64], key0[127:64] };
  assign key3 = { key0[ 63: 0], key1[127:64] };

  assign key_addX = (sel_round_key256 == 1)? key1:
                    (sel_round_key192a == 1)? key2:
                    (sel_round_key192b == 1)? key3:key0;

  assign key_add = key_addX;

endmodule


// ===============================================================================
// Encryption Unit
// ===============================================================================
module EncUnit( data, key_add, final_round, data_next );
  input  [127:0] data;                     // Encryption data input
  input  [127:0] key_add;
  input          final_round;
  output [127:0] data_next;

  wire   [127:0] sRow_out;
  wire   [127:0] sbox_out;
  wire   [127:0] mixcol_out;
  wire   [127:0] enc_result;

// ----------------------------------------------------------------
// Equetions
// ----------------------------------------------------------------
  SubBytes SB3 (data[127:96], sbox_out[127:96]);
  SubBytes SB2 (data[ 95:64], sbox_out[ 95:64]);
  SubBytes SB1 (data[ 63:32], sbox_out[ 63:32]);
  SubBytes SB0 (data[ 31: 0], sbox_out[ 31: 0]);
  ShiftRows SHR( sbox_out, sRow_out );

  MixColumns MX3(sRow_out[127:96], mixcol_out[127:96]);
  MixColumns MX2(sRow_out[ 95:64], mixcol_out[ 95:64]);
  MixColumns MX1(sRow_out[ 63:32], mixcol_out[ 63:32]);
  MixColumns MX0(sRow_out[ 31: 0], mixcol_out[ 31: 0]);

  assign enc_result = (final_round == 0)? mixcol_out : sRow_out;
  assign data_next = enc_result ^ key_add;

endmodule

// ===============================================================================
// Decryotion Unit
// ===============================================================================
module DecUnit( data, key_add, /*count,*/round1, data_next );
  input  [127:0] data;                     // Decryotion data input
  input  [127:0] key_add;
//  input   [14:0] count;
  input          round1;
  output [127:0] data_next;

  wire   [127:0] sRowi_out;
  wire   [127:0] dx;
  wire   [127:0] sboxi_out;
  wire   [127:0] mixi_out;

// ----------------------------------------------------------------
// Equetions
// ----------------------------------------------------------------
  InvMixColumns MXI3(data[127:96], mixi_out[127:96]);
  InvMixColumns MXI2(data[ 95:64], mixi_out[ 95:64]);
  InvMixColumns MXI1(data[ 63:32], mixi_out[ 63:32]);
  InvMixColumns MXI0(data[ 31: 0], mixi_out[ 31: 0]);

//  assign dx = (count[1] == 1)? data: mixi_out;
  assign dx = ( round1 == 1'b1 )? data: mixi_out;
  InvShiftRows SHRI( dx, sRowi_out );

  InvSubBytes ISB3 (sRowi_out[127:96], sboxi_out[127:96]);
  InvSubBytes ISB2 (sRowi_out[ 95:64], sboxi_out[ 95:64]);
  InvSubBytes ISB1 (sRowi_out[ 63:32], sboxi_out[ 63:32]);
  InvSubBytes ISB0 (sRowi_out[ 31: 0], sboxi_out[ 31: 0]);

  assign data_next = sboxi_out ^ key_add;

endmodule


