`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    23:23:51 02/02/2010 
// Design Name: 
// Module Name:    CMAC_TOP 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
`timescale 1ns/1ns

`define  T_IDLE    3'h0
`define  T_KEYGEN  3'h1
`define  T_SUBKEY  3'h2
`define  T_CMAC    3'h3


module CMAC( RSTn, CLK, EN, Kwidth, Kin, Din, Lin, Krdy, Drdy, Lrdy, Kvld, Dvld, Dout, Busy );

// ================================================================================
// Input/Output signals
// ================================================================================
  input           RSTn;
  input           CLK;
  input           EN;
  input     [1:0] Kwidth;            // Key select 00:128bit 01:192bit 10:256bit
  input   [127:0] Kin;
  input   [127:0] Din;
  input    [31:0] Lin;
  input           Krdy;
  input           Drdy;
  input           Lrdy;
  output reg      Kvld;
  output reg      Dvld;
  output  [127:0] Dout;
  output          Busy;

// ================================================================================
// Internal signals
// ================================================================================
  reg      [2:0] state_reg;
  reg             key_start;
  reg             data_start;
  reg             M1_flag;
  reg    [127:0] din;
  reg     [31:0] length_reg;
  wire            key_val;
  wire            data_val;
  wire            mlen0;
  wire            mlen128;
  wire            mlast;

  wire   [127:0] K1_temp1, K1_temp2, K1_temp3;
  wire   [127:0] K2_temp1, K2_temp2, K2_temp3;
  reg    [127:0] K1, K2;
  wire   [127:0] MSB_len;

// =================================================================================================
// Equetions
// =================================================================================================

  ENC_TOP ENC( .RSTn( RSTn ), .CLK( CLK ),
                .enable( EN ), .Kwidth( Kwidth ), .DinKin( din ), .Drdy( data_start ), .Krdy( key_start ),
                .Dout( Dout ), .Busy( Busy ), .Dvld( data_val ), .Kvld( key_val )
              );

  // SUB KEY generation
  assign K1_temp1 = Dout << 1;
  assign K1_temp2 = K1_temp1 ^ 128'h00000000_00000000_00000000_00000087;
  assign K1_temp3 = ( Dout[127] == 1'b1 )? K1_temp2 : K1_temp1;

  assign K2_temp1 = K1_temp3 << 1;
  assign K2_temp2 = K2_temp1 ^ 128'h00000000_00000000_00000000_00000087;
  assign K2_temp3 = ( K1_temp3[127] == 1'b1 )? K2_temp2 : K2_temp1;

  assign mlen0   = ( length_reg[31:0] <= 32'h00000000 )? 1'b1 : 1'b0;
  assign mlen128 = ( length_reg[31:0] <= 32'h00000010 )? 1'b1 : 1'b0;
  assign mlast   = ( length_reg[31:0] <= 32'h00000010 )? 1'b1 : 1'b0;
  assign MSB_len = ( Din & imask( length_reg[4:0] )) ^ padding( length_reg[4:0] );


  // State machine
  always @( posedge CLK or negedge RSTn ) begin
    if ( RSTn == 1'b0 ) begin
      key_start  <= 1'b0;
      M1_flag    <= 1'b0;
      length_reg <= 32'h0000_0000;
      data_start <= 1'b0;
      din <= 128'h00000000_00000000_00000000_00000000;
      K1  <= 128'h00000000_00000000_00000000_00000000;
      K2  <= 128'h00000000_00000000_00000000_00000000;
      Kvld       <= 1'b0;
      Dvld       <= 1'b0;
      state_reg  <= `T_IDLE;
    end
    else if ( EN == 1'b1 ) begin
      case ( state_reg )
        `T_IDLE  : begin
                      if ( Krdy == 1'b1 ) begin
                        key_start <= 1'b1;
                        din <= Kin;
                        Kvld <= 1'b0;
                        state_reg <= `T_KEYGEN;
                      end
                      else if ( Drdy == 1'b1 ) begin
                        if ( M1_flag == 1'b1 )
                          if ( mlen0 == 1'b1 ) din <= K2 ^ padding( 0 );
                          else if ( mlen128 == 1'b1 ) din <= Din ^ padding( length_reg[4:0] ) ^ K1;
                          else din <= Din;
                        else if ( mlast == 1'b1 )
                          if ( length_reg[3:0] != 4'h0 ) din <= MSB_len ^ Dout ^ K2;
                          else din <= Din ^ Dout ^ K1;
                        else din <= Din ^ Dout;

                        Dvld      <= 1'b0;
                        data_start <= 1'b1;
                        state_reg <= `T_CMAC;
                      end
                      else if ( Lrdy == 1'b1 ) begin
                        M1_flag <= 1'b1;
                        length_reg <= Lin[31:0];
                      end
                    end
        `T_KEYGEN : begin
                      if ( Krdy == 1'b1 ) begin
                        key_start <= 1'b1;
                        din <= Kin;
                      end
                      else if (( key_val == 1'b1 ) && ( key_start == 1'b0 )) begin
                        data_start <= 1'b1;
                        din <= 128'h00000000_00000000_00000000_00000000;
                        state_reg <= `T_SUBKEY;
                      end
                      else key_start <= 1'b0;
                    end
        `T_SUBKEY : begin
                      data_start <= 1'b0;

                      if ( data_val == 1'b1 ) begin
                        Kvld <= 1'b1;
                        K1   <= K1_temp3;
                        K2   <= K2_temp3;
                        state_reg <= `T_IDLE;
                      end
                    end
        `T_CMAC   : begin
                      data_start <= 1'b0;

                      if ( data_val == 1'b1 ) begin
                        Dvld    <= 1'b1;
                        M1_flag <= 1'b0;
                        length_reg <= length_reg - 5'h10;
                        state_reg <= `T_IDLE;
                      end
                    end
      endcase
    end
  end


  function [127:0] imask;
  input      [4:0] mlen;
    case ( mlen )
      5'h00: imask = 128'h00000000_00000000_00000000_00000000;
      5'h01: imask = 128'hff000000_00000000_00000000_00000000;
      5'h02: imask = 128'hffff0000_00000000_00000000_00000000;
      5'h03: imask = 128'hffffff00_00000000_00000000_00000000;
      5'h04: imask = 128'hffffffff_00000000_00000000_00000000;
      5'h05: imask = 128'hffffffff_ff000000_00000000_00000000;
      5'h06: imask = 128'hffffffff_ffff0000_00000000_00000000;
      5'h07: imask = 128'hffffffff_ffffff00_00000000_00000000;
      5'h08: imask = 128'hffffffff_ffffffff_00000000_00000000;
      5'h09: imask = 128'hffffffff_ffffffff_ff000000_00000000;
      5'h0a: imask = 128'hffffffff_ffffffff_ffff0000_00000000;
      5'h0b: imask = 128'hffffffff_ffffffff_ffffff00_00000000;
      5'h0c: imask = 128'hffffffff_ffffffff_ffffffff_00000000;
      5'h0d: imask = 128'hffffffff_ffffffff_ffffffff_ff000000;
      5'h0e: imask = 128'hffffffff_ffffffff_ffffffff_ffff0000;
      5'h0f: imask = 128'hffffffff_ffffffff_ffffffff_ffffff00;
      5'h10: imask = 128'hffffffff_ffffffff_ffffffff_ffffffff;
    default: imask = 128'hffffffff_ffffffff_ffffffff_ffffffff;
    endcase
  endfunction

  function [127:0] padding;
  input      [4:0] mlen;
    case ( mlen )
      5'h00: padding = 128'h80000000_00000000_00000000_00000000;
      5'h01: padding = 128'h00800000_00000000_00000000_00000000;
      5'h02: padding = 128'h00008000_00000000_00000000_00000000;
      5'h03: padding = 128'h00000080_00000000_00000000_00000000;
      5'h04: padding = 128'h00000000_80000000_00000000_00000000;
      5'h05: padding = 128'h00000000_00800000_00000000_00000000;
      5'h06: padding = 128'h00000000_00008000_00000000_00000000;
      5'h07: padding = 128'h00000000_00000080_00000000_00000000;
      5'h08: padding = 128'h00000000_00000000_80000000_00000000;
      5'h09: padding = 128'h00000000_00000000_00800000_00000000;
      5'h0a: padding = 128'h00000000_00000000_00008000_00000000;
      5'h0b: padding = 128'h00000000_00000000_00000080_00000000;
      5'h0c: padding = 128'h00000000_00000000_00000000_80000000;
      5'h0d: padding = 128'h00000000_00000000_00000000_00800000;
      5'h0e: padding = 128'h00000000_00000000_00000000_00008000;
      5'h0f: padding = 128'h00000000_00000000_00000000_00000080;
    default: padding = 128'h00000000_00000000_00000000_00000000;
    endcase
  endfunction

endmodule
