/*-------------------------------------------------------------------------
 File name   : FPGA1_CMAC.v
 Version     : Version 1.0
 Created     : 
 Last update : 
 Desgined by : 
 
 
 Copyright (C) 2010 AIST
 
 -------------------------------------------------------------------------*/ 


//================================================
`timescale 1ns / 1ps


module FPGA1_SHA256
  (// Local bus II
   lbus_rstn, lbus_clk,
   lbus_wd,  lbus_we, lbus_ful, lbus_aful,
   lbus_rd, lbus_re, lbus_emp, lbus_aemp,
   lbus_rdy, 
   
   trig_startn, trig_endn, trig_exec, // Trigger
   led); // LED.
   
   //------------------------------------------------
   // Local bus II
   input [7:0]  lbus_wd;                 // write data input
   input        lbus_we;
   output       lbus_ful, lbus_aful;
   output [7:0] lbus_rd;                 // read data output
   input        lbus_re;
   output       lbus_emp, lbus_aemp;
   output       lbus_rdy;
   input        lbus_clk, lbus_rstn;

   output       trig_startn, trig_endn, trig_exec; // Trigger
   output [7:0] led;           // LED
   
   //------------------------------------------------
   // Internal clock
   wire         rst;
   wire         resetn;
   wire         clk;

   wire [7:0]   wd, rd;
   wire         we, ful, aful, re, emp, aemp;
   wire [11:0]  rcnt;

   wire [15:0]  lbus_a, lbus_di, lbus_do;
   wire         lbus_w, lbus_r;

   // Block SHA256
   wire   [31:0] blk_din;
   wire  [255:0] blk_dout;
   wire          blk_krdy, blk_drdy, blk_dvld;
   wire          blk_en, blk_rstn, blk_busy;
  
   //------------------------------------------------
   assign led[0] = ~rst;
   assign led[1] = ~blk_rstn;
   assign led[2] = 1'b1;
   assign led[3] = ~blk_drdy;
   assign led[4] = ~blk_drdy;
   assign led[5] = ~blk_drdy;
   assign led[6] = ~blk_dvld;
   assign led[7] = ~blk_busy;

   assign lbus_rdy = ~rst;
   
   assign trig_startn = ~blk_drdy;
   assign trig_endn   = ~blk_dvld;
   assign trig_exec   = blk_busy;

  assign resetn = blk_rstn & ~rst;
   
   //------------------------------------------------   
   SYNCFIFO_8x2047 fifo_w(
      .wd(lbus_wd), .we(lbus_we), .ful(lbus_ful), .aful(lbus_aful),
      .rd(rd), .re(re), .emp(emp), .aemp(aemp),
      .cnt(rcnt), .clk(clk), .rst(rst)
     );

   SYNCFIFO_8x2047 fifo_r(
      .wd(wd), .we(we), .ful(ful), .aful(aful),
      .rd(lbus_rd), .re(lbus_re), .emp(lbus_emp), .aemp(lbus_aemp),
      .cnt(), .clk(clk), .rst(rst)
     );

   CTRL_LBUS ctrl_lbus
     (.lbus_a(lbus_a), .lbus_di(lbus_di), .lbus_do(lbus_do),
      .lbus_wr(lbus_w), .lbus_rd(lbus_r),
      .wd(wd), .we(we), .ful(ful), .aful(aful),
      .rd(rd), .re(re), .emp(emp), .aemp(aemp), .rcnt(rcnt),
      .clk(clk), .rst(rst));

   LBUS_IF lbus_if
     (.lbus_a(lbus_a), .lbus_di(lbus_di), .lbus_do(lbus_do),
      .lbus_wr(lbus_w), .lbus_rd(lbus_r),
      .blk_din(blk_din), .blk_dout(blk_dout),
      .blk_krdy(blk_krdy), .blk_drdy(blk_drdy), 
      .blk_dvld(blk_dvld),
      .blk_en(blk_en), .blk_rstn(blk_rstn),
      .clk(clk), .rst(rst));

   //------------------------------------------------   

   SHA256 SHA( .RSTn( resetn ), .CLK( clk ),
               .EN( blk_en ), .INIT( blk_krdy ), .MSGin( blk_din[31:0] ), .Mrdy( blk_drdy ),
               .Hout( blk_dout ), .Hvld( blk_dvld ), .Busy( blk_busy )
             );

   //------------------------------------------------   
   MK_CLKRST mk_clkrst (.clkin(lbus_clk), .rstnin(lbus_rstn),
                        .clk(clk), .rst(rst));
endmodule // CHIP_SASEBO_GII_LX50_SHA256


   
//================================================ MK_CLKRST
module MK_CLKRST (clkin, rstnin, clk, rst);
   //synthesis attribute keep_hierarchy of MK_CLKRST is no;
   
   //------------------------------------------------
   input  clkin, rstnin;
   output clk, rst;
   
   //------------------------------------------------
   wire   refclk;
   wire   clk_dcm, locked;

   //------------------------------------------------ clock
   IBUFG u10 (.I(clkin), .O(refclk)); 
   
   DCM_BASE u11 (.CLKIN(~refclk), .CLKFB(clk), .RST(~rstnin),
                 .CLK0(clk_dcm),     .CLKDV(),
                 .CLK90(), .CLK180(), .CLK270(),
                 .CLK2X(), .CLK2X180(), .CLKFX(), .CLKFX180(),
                 .LOCKED(locked));
 
   BUFG  u12 (.I(clk_dcm),   .O(clk));

   //------------------------------------------------ reset
   MK_RST_BUFG u20 (.locked(locked), .clk(clk), .rst(rst));
endmodule // MK_CLKRST



//================================================ MK_RST_BUFG
module MK_RST_BUFG (locked, clk, rst);
   //synthesis attribute keep_hierarchy of MK_RST_BUFG is no;
   
   //------------------------------------------------
   input  locked, clk;
   output rst;

   //------------------------------------------------
   reg [15:0] rst_reg;
   wire       rst0;
   
   //------------------------------------------------
   always @(posedge clk or negedge locked) 
     if (~locked) rst_reg <= 16'hFFFF;
     else         rst_reg <= {1'b0,rst_reg[15:1]};

   assign rst0 = rst_reg[0];
   //synthesis attribute maxdelay of rstn is 1ns;
   
   BUFG u0 (.I(rst0), .O(rst));
endmodule // MK_RST_BUFG
